// $Id: avrdomd-no-socket.c,v 1.1 2010/08/18 12:06:34 swaj Exp $

// AVRDOM logger daemon

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <time.h>
#include <ctype.h>
#include <signal.h>

#define BAUDRATE B9600
//#define BAUDRATE B2400

static uint8_t foreground=0;
static uint8_t say_tty=1;

#define LOG_BUFFER_MAX 128
static char log_buffer[LOG_BUFFER_MAX+1];

#define PROG_NAME "avrdomd"
#define LOCK_DIR "/var/lock"
#define DEVICE_MAX 64

static FILE *fl = NULL;
static int fd=-1;
static char lock_name[DEVICE_MAX+1];
static int lock_created=0;

//--------------------------------- help page  ---------------------------------
void print_help (void) 
{
	printf ("\n");
	printf ("Usage: avrdomd -d /dev/ttyS?\n");
	printf ("               -f foreground\n");
	exit (0);
}

// ------------------- logging -------------------------------------------------
static
void say (const char *fmt, ...)
{
	va_list ptr;                                                                                         
	va_start(ptr, fmt);

	vsnprintf (log_buffer, LOG_BUFFER_MAX, fmt, ptr);

        va_end(ptr);

	syslog(LOG_INFO, "%s", log_buffer);
	if (say_tty)
		printf ("%s", log_buffer);
}

static void remove_lock_file (void);

// ------------------- signals -------------------------------------------------
static
void clean_quit (int signo)
{
	say ("Got termination signal %d, quit\n", signo);
	if (fl)
		fclose(fl);
	
	if (fd >=0)
		close (fd);

	remove_lock_file ();

	exit (0);
}

// ----------------- locking ---------------------------------------------------
static
int create_lock_file(char * dev)
{
	char *ptr;
	FILE *f;

	ptr = strrchr(dev, '/');

	if (! ptr) {
		say ("probably wront tty name %s (no '/')\n", dev);
		return 1;
	}
	snprintf (lock_name, DEVICE_MAX, "%s/LCK..%s", LOCK_DIR, ++ptr);

	say ("'%s'\n", lock_name);

	f=fopen (lock_name, "r");

	if (f) {
		lock_name[0] = 0x0;
		ptr = fgets(lock_name, DEVICE_MAX, f);
		fclose (f); f=0;
		if (ptr)
			ptr[strlen(ptr)-1]= 0x0;
	
		say("Lock already exist for %s [%s ]\n", dev, lock_name);
		return 1;
	}
	
	f = fopen (lock_name, "a");

	if (! f)
	{
		say("Cannot create lock file %s : %s\n", lock_name, strerror(errno) );
		return 1;
	}

	fprintf (f, "      %d %s %s\n", getpid(), PROG_NAME,
		 (getenv("USER")) ? getenv("USER") : "noone");

	fclose (f);
	f = 0;
	lock_created=1;

	return 0;
}

static
void remove_lock_file (void)
{
	if (!lock_created)
		return;

	if ( remove(lock_name))
	{
		say("Cannot remove lock file: %s\n", strerror(errno) ); 
	}
}



//--------------------------------- MAIN ---------------------------------------
int main (int argc, char * argv[], char *env[] )
{
	char ch;
	struct termios oldtio,newtio;
	char devfile[DEVICE_MAX+1];

#define BUF_SIZE 80	
	char buf[BUF_SIZE+1];


//--------------------------------------------  . 
	foreground=0;

	openlog(PROG_NAME, LOG_PID, LOG_LOCAL7);                                                              
	say("%s started\n", PROG_NAME);

	if (argc <3)
		print_help();

	while ( (ch=getopt( argc, argv, "d:f") ) != EOF )
	{
		switch (ch)
		{
		case 'd':
			strncpy ( devfile, optarg, 63);
			break; 
			
		case 'f':
			foreground=1;
			break; 
			
		default:
			print_help();
			break;
		}
	}

//--------------------------------------------   /dev/tty...

   say ("Opening terminal '%s'\n", devfile); 

   if (create_lock_file(devfile))
   {
	   say ("Cannot create lock file, exit\n");
	   exit (-1);
   }

   fd = open(devfile, O_RDWR );

   if (fd <0) 
   {
	   say ("Cannot open terminal: %s\n", strerror(errno) );
	   remove_lock_file ();
	   exit(-1);
   }


   tcgetattr(fd,&oldtio);
   bzero( (void*)&newtio, sizeof(newtio));

   newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
   newtio.c_iflag = IGNPAR | ICRNL;
   newtio.c_oflag = 0;
   newtio.c_lflag = ICANON;

   newtio.c_cc[VINTR]    = 0;     /* Ctrl-c */
   newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
   newtio.c_cc[VERASE]   = 0;     /* del */
   newtio.c_cc[VKILL]    = 0;     /* @ */
   newtio.c_cc[VEOF]     = 4;     /* Ctrl-d */
   newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
   newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives */
   newtio.c_cc[VSWTC]    = 0;     /* '\0' */
   newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */
   newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
   newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
   newtio.c_cc[VEOL]     = 0;     /* '\0' */
   newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
   newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
   newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
   newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
   newtio.c_cc[VEOL2]    = 0;     /* '\0' */

   tcflush(fd, TCIFLUSH);
   tcsetattr(fd,TCSANOW,&newtio);

   fl = fdopen(fd, "r");

   if (!fl) {
	   say ("fdopen() failed: %s\n", strerror(errno) );
	   remove_lock_file ();
	   exit(-1);
   }

   say ("Terminal is ready\n"); 

   if (! foreground )
   {
	   int pid=0;

	   say("Going daemon mode\n");

	   pid = fork();

	   switch (pid)
	   {

	   case 0: /* child */
		   say_tty=0;
		   
		   chdir("/");
		   close (STDIN_FILENO);
		   close (STDOUT_FILENO);
		   close (STDERR_FILENO);
		   setsid();
		   umask(0);
		   
		   say ("Daemon mode Ok\n");
		   break;
	   case -1:
		   say ("can't fork: %s", strerror(errno));
		   remove_lock_file ();
		   exit (-1);

	   default:
		   say ("Process forked, pid %d\n", pid);
		   
		   if (fl)
			   fclose(fl);
	
		   if (fd >=0)
			   close (fd);
		   
		   exit (0);
	   }
   }

   signal(SIGINT,(void*)&clean_quit);                                                      
   signal(SIGTERM,(void*)&clean_quit);                                                     
   signal(SIGQUIT,(void*)&clean_quit);      

// ----------------------------------------- main loop (recieving mesages ------
   while (1)
   {
	   if (fgets( buf, BUF_SIZE, fl))
	   {
	   	   if (strlen(buf) > 5)
		   {
			   if ( ! strncmp(buf, "LINF", 4) ||
				! strncmp(buf, "LERR", 4) ||
				! strncmp(buf, "LDBG", 4) )
			   {
				   buf[strlen(buf)-1]= 0x0;
				   say ("AVRDOM:%s\n", buf);
			   }
		   }
	   }
   }
}


// __END__
